from __future__ import annotations

from typing import TYPE_CHECKING

from networkx.algorithms.isomorphism import DiGraphMatcher

if TYPE_CHECKING:
    from parchmint import Device


class SimilarityMatcher(DiGraphMatcher):
    """Implementation of VF2 algorithm for matching undirected graphs.
    Suitable for Graph and MultiGraph instances.
    """

    def __init__(
        self,
        G1: Device,
        G2: Device,
        # semantic_information: Dict[str, NodeFilter],
        compare_params=False,
        check_connection_target=False,
    ):
        # self._semantic_information = semantic_information
        self._G1_device = G1
        self._G2_device = G2
        self._param_flag = compare_params
        self._connection_flag = check_connection_target
        self._G1_in_edges = G1.G.in_edges
        self._G1_out_edges = G1.G.out_edges
        self._G2_in_edges = G2.G.in_edges
        self._G2_out_edges = G2.G.out_edges
        self._G1_in_edges_diff_list = []
        self._G1_out_edges_diff_list = []
        self._G2_in_edges_diff_list = []
        self._G2_out_edges_diff_list = []
        self._G1_param_diff_list = []
        self._G2_param_diff_list = []
        self._G1_layer_diff_list = []
        self._G2_layer_diff_list = []
        self._G1_port_diff_list = []
        self._G2_port_diff_list = []

        super(SimilarityMatcher, self).__init__(G1.G, G2.G)

    def semantic_feasibility(self, G1_node, G2_node) -> bool:
        """Overriding semantic_feasibility to compare the layers, params, ports, and connections

        Args:
            G1_node (str): node 1
            G2_node (str): node 2

        Returns:
            bool: if they are semantically feasible, return true. else return false.
        """
        feasible = True

        G1_component = self._G1_device.get_component(G1_node)
        G2_component = self._G2_device.get_component(G2_node)

        # compare connectivities
        for item in self._G1_in_edges:
            conn_g1 = self._G1_in_edges[item]
            conn_g2 = self._G2_in_edges[item]

            g1_source = conn_g1["source_port"]
            g2_source = conn_g2["source_port"]

            g1_sink = conn_g1["sink_port"]
            g2_sink = conn_g2["sink_port"]

            if (g1_source.component != g2_source.component) or (
                g1_source.port != g2_source.port
            ):
                print("source port wrong")
                self._G1_in_edges_diff_list.append(g1_source.component)
                self._G1_in_edges_diff_list.append(g1_source.port)
                self._G2_in_edges_diff_list.append(g2_source.component)
                self._G2_in_edges_diff_list.append(g2_source.port)
                if self._connection_flag:
                    feasible = False

            if (g1_sink.component != g2_sink.component) or (
                g1_sink.port != g2_sink.port
            ):
                print("sink port wrong")
                self._G1_out_edges_diff_list.append(g1_sink.component)
                self._G1_out_edges_diff_list.append(g1_sink.port)
                self._G2_out_edges_diff_list.append(g2_sink.component)
                self._G2_out_edges_diff_list.append(g2_sink.port)
                if self._connection_flag:
                    feasible = False

        # compare layers
        if G1_component.layers != G2_component.layers:
            print("layer wrong")
            self._G1_layer_diff_list.append(G1_component.layers)
            self._G2_layer_diff_list.append(G2_component.layers)
            feasible = False

        # compare params
        if G1_component.params != G2_component.params:
            print("params wrong")
            self._G1_param_diff_list.append(G1_component.params)
            self._G2_param_diff_list.append(G2_component.params)

            if self._param_flag is True:
                feasible = False
            else:
                feasible = True

        # compare ports
        if G1_component.ports != G2_component.ports:
            print("ports wrong")
            self._G1_port_diff_list.append(G1_component.ports)
            self._G2_port_diff_list.append(G2_component.ports)
            feasible = False

        return feasible

    def print_params_diff(self) -> None:
        """
        This method prints out the difference in the parameters between G1 and G2
        """
        print("----Param differences----")

        for i in range(len(self._G1_param_diff_list)):
            print(
                f"G1: {self._G1_param_diff_list[i]}, G2: {self._G2_param_diff_list[i]}"
            )

        print("----End----")

    def print_layers_diff(self) -> None:
        """
        This method prints out the difference in the layers between G1 and G2
        """
        print("----Layer differences----")

        for i in range(len(self._G1_layer_diff_list)):
            print(
                f"G1: {self._G1_layer_diff_list[i]}, G2: {self._G2_layer_diff_list[i]}"
            )

        print("----End----")

    def print_port_diff(self) -> None:
        """
        This method prints out the difference in the ports between G1 and G2
        """
        print("----Port differences----")

        for i in range(len(self._G1_port_diff_list)):
            print(f"G1: {self._G1_port_diff_list[i]}, G2: {self._G2_port_diff_list[i]}")

        print("----End----")

    def print_in_edges_diff(self) -> None:
        """
        This method prints out the difference in the in edges between G1 and G2
        """
        print("----In edges differences----")

        for i in range(0, len(self._G1_in_edges_diff_list), 2):
            print(
                f"G1: {self._G1_in_edges_diff_list[i]} -port: {self._G1_in_edges_diff_list[i+1]}, G2: {self._G2_in_edges_diff_list[i]} -port: {self._G2_in_edges_diff_list[i+1]}"
            )

        print("----End----")

    def print_out_edges_diff(self) -> None:
        """
        This method prints out the difference in the out edges between G1 and G2
        """
        print("----Out edges differences----")

        for i in range(0, len(self._G1_out_edges_diff_list), 2):
            print(
                f"G1: {self._G1_out_edges_diff_list[i]} -port: {self._G1_out_edges_diff_list[i+1]}, G2: {self._G2_out_edges_diff_list[i]} -port: {self._G2_out_edges_diff_list[i+1]}"
            )

        print("----End----")



from __future__ import annotations

import json
import pathlib
from enum import Enum
from typing import Dict, List, Optional

import jsonschema
import networkx as nx

import parchmint
from parchmint.component import Component
from parchmint.connection import Connection
from parchmint.layer import Layer
from parchmint.params import Params
from parchmint.similaritymatcher import SimilarityMatcher

PROJECT_DIR = pathlib.Path(parchmint.__file__).parent.parent.absolute()

# GM = nx.algorithms.isomorphism.GraphMatcher(device1, device2)


class ValveType(Enum):
    NORMALLY_OPEN = 0
    NORMALLY_CLOSED = 1

    def __str__(self) -> str:
        if self == ValveType.NORMALLY_OPEN:
            return "NORMALLY_OPEN"
        elif self == ValveType.NORMALLY_CLOSED:
            return "NORMALLY_CLOSED"
        else:
            raise Exception("Could not generate Valve Type string")

    def __eq__(self, o: object) -> bool:
        if o.__class__ is ValveType:
            return super().__eq__(o)
        elif o.__class__ is str:
            if self is ValveType.NORMALLY_OPEN and o == "NORMALLY_OPEN":
                return True
            elif self is ValveType.NORMALLY_CLOSED and o == "NORMALLY_CLOSED":
                return True
            else:
                return False
        else:
            return False


class Device:
    def __init__(self, json_data=None):
        """Creates a new device object

        Args:
            json (dict, optional): json dict after json.loads(). Defaults to None.
        """
        self.name: str = ""
        self.components: List[Component] = []
        self.connections: List[Connection] = []
        self.layers: List[Layer] = []
        self.params: Params = Params()
        self.features = []  # Store Raw JSON Objects for now
        self.xspan: Optional[int] = None
        self.yspan: Optional[int] = None
        self.G = nx.MultiDiGraph()

        # Stores the valve / connection mappings
        self._valve_map: Dict[Component, Connection] = {}
        self._valve_type_map: Dict[Component, ValveType] = {}

        if json_data:
            self.parse_from_json(json_data)
            self.generate_network()

    def map_valve(
        self,
        valve: Component,
        connection: Connection,
        type_info: Optional[ValveType] = None,
    ) -> None:
        """Maps the valve to a connection in the device

        Args:
            valve (Component): valve component
            connection (Connection): connection on which the valve is mapped
            type_info (Optional[ValveType]): Type informaiton of the valve

        """
        self._valve_map[valve] = connection
        if type_info is not None:
            self.update_valve_type(valve, type_info)

    def get_valves(self) -> List[Component]:
        """Returns the list of valves in the device

        Returns:
            List[Component]: Valve Component Objects
        """
        return list(self._valve_map.keys())

    def get_valve_connection(self, valve: Component) -> Connection:
        """Returns the connection associated with the valve object

        Args:
            valve (Component): Valve object for which we are finding the connection

        Returns:
            Connection: connection object on which the valve is placed
        """
        return self._valve_map[valve]

    def update_valve_type(self, valve: Component, type_info: ValveType) -> None:
        """Updates the type of the valve to normally closed  or normally open

        Args:
            valve (Component): Valve object we want to update
            type_info (ValveType): Valve Type

        Raises:
            KeyError: Raises the error if the valve object is not mapped as a valve in the device
        """
        if valve in list(self._valve_map.keys()):
            self._valve_type_map[valve] = type_info
        else:
            raise KeyError(
                "Could not update type for valve: {} since it is not found in the valveMap of they device".format(
                    valve.ID
                )
            )

    # add compare function
    # -pass device, compare devices
    # - check connections, components, print if its same or not
    # - have a flag to print parameter differences

    def compare(self, device: Device) -> bool:
        """compare against the input device. Return true if they are semnatcally feasible.

        Args:
            device (Device): expected device

        Returns:
            bool: If semntically feasible, return true. Else false.
        """
        self.generate_network()

        SM = SimilarityMatcher(self, device)

        is_same = SM.is_isomorphic()
        SM.print_params_diff()
        SM.print_layers_diff()
        SM.print_port_diff()
        SM.print_in_edges_diff()
        SM.print_out_edges_diff()

        if is_same:
            print("Match!")
        else:
            print("Not Match!")

        return is_same

    def add_component(self, component: Component) -> None:
        """Adds a component object to the device

        Args:
            component (Component): component to eb added

        Raises:
            Exception: if the passed object is not a Component instance
        """
        if isinstance(component, Component):
            # Check if Component Exists, if it does ignore it
            if self.component_exists(component.ID):
                print(
                    "Component {} already present in device, "
                    "hence skipping the component".format(component.name)
                )
            self.components.append(component)
        else:
            raise Exception(
                "Could not add component since its not an instance of parchmint:Component"
            )

    def add_connection(self, connection: Connection) -> None:
        """Adds a connection object to the device

        Args:
            connection (Connection): connectin to add

        Raises:
            Exception: if the arg is not a Connection type object
        """
        if isinstance(connection, Connection):
            self.connections.append(connection)
        else:
            raise Exception(
                "Could not add component since its not an instance of parchmint:Connection"
            )

    def add_layer(self, layer: Layer) -> None:
        """Adds a layer to the device

        Args:
            layer (Layer): layer to be added to the device
        """
        if isinstance(layer, Layer):
            self.layers.append(layer)

    def get_layer(self, id: str) -> Layer:
        """Returns the layer with the corresponding id

        Args:
            id (str): id of the layer

        Raises:
            Exception: if a layer with the corresponding id is not present

        Returns:
            Layer: layer with the corresponding id
        """
        for layer in self.layers:
            if layer.ID == id:
                return layer
        raise Exception("Could not find the layer {}".format(id))

    def merge_netlist(self, netlist: Device) -> None:
        """Merges two netlists together. Currently assumes that both
        devices have the same ordering of layers

        Args:
            netlist (Device): netlist to merge
        """
        # TODO - Figure out how to merge the layers later
        # First create a map of layers
        layer_mapping = {}
        for layer in netlist.layers:
            if layer not in self.layers:
                self.add_layer(layer)
                layer_mapping[layer] = layer
            else:
                assert layer.ID is not None
                layer_mapping[layer] = self.get_layer(layer.ID)

        for component in netlist.components:
            new_layers = []
            for layer in component.layers:
                new_layers.append(layer_mapping[layer])
            component.layers = new_layers
            self.add_component(component)

        for connection in netlist.connections:
            connection.layer = layer_mapping[connection.layer]
            self.add_connection(connection)

    def parse_from_json(self, json_data) -> None:
        """Returns the json dict

        Returns:
            dict: dictionary that can be used in json.dumps()
        """
        self.name = json_data["name"]

        # First always add the layers
        if "layers" in json_data.keys():
            for layer in json_data["layers"]:
                self.add_layer(Layer(layer))
        else:
            print("no layers found")

        # Loop through the components
        if "components" in json_data.keys():
            for component in json_data["components"]:
                self.add_component(Component(component, self))
        else:
            print("no components found")

        if "connections" in json_data.keys():
            for connection in json_data["connections"]:
                self.add_connection(Connection(connection, self))
        else:
            print("no connections found")

        if "params" in json_data.keys():
            self.params = Params(json_data["params"])

            if self.params.exists("xspan"):
                self.xspan = self.params.get_param("xspan")
            elif self.params.exists("width"):
                self.xspan = self.params.get_param("width")
            elif self.params.exists("x-span"):
                self.xspan = self.params.get_param("x-span")

            if self.params.exists("yspan"):
                self.yspan = self.params.get_param("yspan")
            elif self.params.exists("length"):
                self.yspan = self.params.get_param("length")
            elif self.params.exists("y-span"):
                self.yspan = self.params.get_param("y-span")
        else:
            print("no params found")

        if "valveMap" in json_data.keys():
            valve_map = json_data["valveMap"]

            for key, value in valve_map.items():
                self._valve_map[self.get_component(key)] = self.get_connection(value)

        if "valveTypeMap" in json_data.keys():
            valve_type_map = json_data["valveTypeMap"]

            for key, value in valve_type_map.items():
                if value is ValveType.NORMALLY_OPEN:
                    self._valve_type_map[
                        self.get_component(key)
                    ] = ValveType.NORMALLY_OPEN
                else:
                    self._valve_type_map[
                        self.get_component(key)
                    ] = ValveType.NORMALLY_CLOSED

    def get_components(self) -> List[Component]:
        """Returns the components in the device

        Returns:
            List[Component]: list of components in the device
        """
        return self.components

    def get_connections(self) -> List[Connection]:
        """Returns the connections in the device

        Returns:
            List[Connection]: list of connections in the device
        """
        return self.connections

    def generate_network(self) -> None:
        """Generates the underlying graph"""
        for component in self.components:
            self.G.add_node(component.ID, component_ref=component)

        for connection in self.connections:
            sourceref = connection.source.component
            for sink in connection.sinks:
                sinkref = sink.component
                self.G.add_edge(
                    sourceref,
                    sinkref,
                    source_port=connection.source,
                    sink_port=sink,
                    connection_ref=connection,
                )

    def get_name_from_id(self, id: str) -> str:
        """Returns the name of the component with the corresponding id

        Args:
            id (str): id of the object

        Returns:
            Optional[str]: name of the corresponding object
        """
        for component in self.components:
            if component.ID == id:
                return component.name

        raise Exception("Could not find component with ID: {}".format(id))

    def component_exists(self, component_id: str) -> bool:
        """checks if component exists in the device

        Args:
            component_id (str): id of the component

        Returns:
            bool: true if the component exists
        """
        for component in self.components:
            if component_id == component.ID:
                return True

        return False

    def connection_exists(self, connection_id: str) -> bool:
        """checks if connection exists in the device

        Args:
            connection_id (str): id of the connection

        Returns:
            bool: true if the connection exists
        """
        for connection in self.connections:
            if connection_id == connection.ID:
                return True

        return False

    def get_component(self, id: str) -> Component:
        """Returns the component with the corresponding ID

        Args:
            id (str): id of the component

        Raises:
            Exception: if the component is not found

        Returns:
            Component: component with the corresponding id
        """
        for component in self.components:
            if component.ID == id:
                return component
        raise Exception("Could not find component with id {}".format(id))

    def get_connection(self, id: str) -> Connection:
        """Returns the connection with the corresponding id

        Args:
            id (str): id of the connection

        Raises:
            Exception: if the connection is not found

        Returns:
            Connection: connection with the corresponding id
        """
        for connection in self.connections:
            if connection.ID == id:
                return connection
        raise Exception("Could not find connection with id {}".format(id))

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self.__dict__)

    def to_parchmint_v1(self):
        """Returns the json dict

        Returns:
            dict: dictionary that can be used in json.dumps()
        """
        ret = {}
        ret["name"] = self.name
        ret["components"] = [c.to_parchmint_v1() for c in self.components]
        ret["connections"] = [c.to_parchmint_v1() for c in self.connections]
        ret["params"] = self.params.to_parchmint_v1()
        ret["layers"] = [layer.to_parchmint_v1() for layer in self.layers]
        ret["version"] = 1

        return ret

    def to_parchmint_v1_x(self) -> Dict:
        ret = self.to_parchmint_v1()

        # Modify the version of the parchmint
        ret["version"] = 1.2

        # Add the valvemap information
        valve_map = {}
        valve_type_map = {}
        for valve, connection in self._valve_map.items():
            valve_map[valve.ID] = connection.ID
        ret["valveMap"] = valve_map
        for valve, valve_type in self._valve_type_map.items():
            valve_type_map[valve.ID] = str(valve_type)
        ret["valveTypeMap"] = valve_type_map
        return ret

    @staticmethod
    def validate_V1(json_str: str) -> None:
        """Validates the json string against the schema

        Args:
            json_str (str): json string
        """
        schema_path = PROJECT_DIR.joinpath("schemas").joinpath("V1.json")
        with open(schema_path) as json_file:
            schema = json.load(json_file)
            json_data = json.loads(json_str)
            validator = jsonschema.Draft7Validator(schema)

            errors = validator.iter_errors(json_data)  # get all validation errors

            for error in errors:
                print(error)
                print("------")



from os import truncate


class Params:
    def __init__(self, json=None):
        """Creates an instance of the params

        Args:
            json (dict, optional): json dict form json.loads(). Defaults to None.
        """
        self.data = {}

        if json:
            self.parse_from_json(json)

    def __ne__(self, other) -> bool:
        """operator overload to compare two params.
        ex. P1 != P2

        Args:
            other (Params): P2 part

        Returns:
            bool: Return true if NOT equal. Otherwise false
        """
        len1 = len(self.data)
        len2 = len(self.data)

        if len1 != len2:
            return True
        else:
            for item in self.data:
                if item not in other.data:
                    return True
                else:
                    if self.data[item] != other.data[item]:
                        return True

        return False

    def get_param(self, key: str):
        """Returns the value stored against the key

        Args:
            key (str): parameter name

        Returns:
            [type]: [description]
        """
        return self.data[key]

    def set_param(self, key: str, value):
        """Sets the value for a given key to the param

        Args:
            key (str): parameter name
            value (): value to be stored against the key
        """
        self.data[key] = value

    def exists(self, key: str) -> bool:
        """Checks if the key exists in the params

        Args:
            key (str): key of the param

        Returns:
            bool: true if key is present in the params
        """
        return key in self.data.keys()

    def parse_from_json(self, json):
        """Parses from the json dict

        Args:
            json (dict): json dict after json.loads()
        """
        for key, value in json.items():
            self.data[key] = value

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self.__dict__)

    def to_parchmint_v1(self) -> dict:
        """Returns the json dict

        Returns:
            dict: dictionary that can be used in json.dumps()
        """
        return self.data



# These are done in a specific format to prevent circular imports
# Think of them as an escalation from base component to most complex
from .component import Component
from .connection import Connection
from .device import Device
from .layer import Layer
from .params import Params
from .port import Port
from .target import Target



from typing import Optional


class Target:
    def __init__(self, json=None):
        """Creates a Target object that describes where the connection will connect to

        Args:
            json (dict, optional): json dict from json.loads(). Defaults to None.
        """
        self._component = ""
        self._port: Optional[str] = None

        if json:
            self.parse_from_json(json)

    def parse_from_json(self, json):
        """Loads the instance data from the json dict

        Args:
            json ([type]): json dict from json.loads()
        """
        self._component = json["component"]
        self._port = json["port"]

    @property
    def component(self) -> str:
        """Returns the component in the Target

        Returns:
            str: Target component ID
        """
        return self._component

    @component.setter
    def component(self, value: str) -> None:
        """Sets the component in the target

        Args:
            value (str): component ID
        """
        self._component = value

    @property
    def port(self) -> Optional[str]:
        """Returns the port label in the target

        Returns:
            Optional[str]: port label of the target
        """
        return self._port

    @port.setter
    def port(self, value: Optional[str]) -> None:
        """Sets the port in the target

        Args:
            value (str): port label for the given component
        """
        self._port = value

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self.__dict__)

    def to_parchmint_v1(self) -> dict:
        """Returns the json dict

        Returns:
            dict: dictionary that can be used in json.dumps()
        """
        return {
            "component": self._component,
            "port": self._port,
        }

    def __eq__(self, obj):
        if isinstance(obj, Target):
            return obj.component == self.component and obj.port == self.port
        else:
            return False



import argparse
from pathlib import Path

from parchmint.device import Device


def validate_V1():
    parser = argparse.ArgumentParser()

    parser.add_argument("input", help="This is the file thats used as the input ")

    args = parser.parse_args()
    file_path = Path(args.input).resolve()
    with open(file_path) as data_file:
        Device.validate_V1(data_file.read())



class Port:
    def __init__(self, json=None):
        """Creates a ComponentPort which is used to represent the points
        where a connection connects on the component

        Args:
            json (dict, optional): json dict. Defaults to None.
        """
        self.x: int = -1
        self.y: int = -1
        self.label: str = ""
        self.layer: str = ""

        if json:
            self.parse_from_json(json)

    def parse_from_json(self, json):
        """Parses the json dict from json.loads()

        Args:
            json ([dict): dictionary
        """
        self.x = json["x"]
        self.y = json["y"]
        self.label = json["label"]
        self.layer = json["layer"]

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self.__dict__)

    def to_parchmint_v1(self):
        """Returns the json dict

        Returns:
            dict: dictionary that can be used in json.dumps()
        """
        return {
            "x": self.x,
            "y": self.y,
            "label": self.label,
            "layer": self.layer,
        }

    def __eq__(self, obj):
        if isinstance(obj, Port):
            return obj.label == self.label
        else:
            return False



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['parchmint']

package_data = \
{'': ['*']}

install_requires = \
['argparse>=1.4.0,<2.0.0', 'jsonschema>=3.2.0,<4.0.0', 'networkx>=2.5,<3.0']

entry_points = \
{'console_scripts': ['parchmint-validate = parchmint.cmdline:validate_V1']}

setup_kwargs = {
    'name': 'parchmint',
    'version': '0.2.10',
    'description': 'Parchmint is data interchange standard for representing microfluidic designs. Check out https://parchmint.org for more information.',
    'long_description': None,
    'author': 'Radhakrishna Sanka',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import List, Optional, Tuple

from parchmint.layer import Layer
from parchmint.params import Params
from parchmint.target import Target


class ConnectionPath:
    def __init__(
        self,
        source: Target = None,
        sink: Target = None,
        waypoints: List[Tuple[int, int]] = None,
        json=None,
    ) -> None:
        """Creates a new connection path object

        Args:
            source (Target): source corresponding to the path
            sink (Target): sink corresponding to the path
            waypoints (List[Tuple[int, int]], optional): list of the coordinates. Defaults to [].
        """
        if waypoints is None:
            waypoints = []
        super().__init__()
        self.__source: Optional[Target] = source
        self.__sink: Optional[Target] = sink
        self.__waypoints: List[Tuple[int, int]] = waypoints

        if json is not None:
            self.parse_parchmint_v1(json)

    @property
    def source(self) -> Target:
        if self.__source is None:
            raise AssertionError
        return self.__source

    @property
    def sink(self) -> Target:
        if self.__sink is None:
            raise AssertionError
        return self.__sink

    @property
    def waypoints(self) -> List[Tuple[int, int]]:
        return self.__waypoints

    @waypoints.setter
    def waypoints(self, value: List[Tuple[int, int]]):
        self.__waypoints = value

    def to_parchmint_v1(self):
        """Returns the json dict

        Returns:
            dict: dictionary that can be used in json.dumps()
        """
        return {
            "source": None
            if self.__source is None
            else self.__source.to_parchmint_v1(),
            "sink": None if self.__sink is None else self.__sink.to_parchmint_v1(),
            "wayPoints": [list(wp) for wp in self.__waypoints],
        }

    def parse_parchmint_v1(self, json) -> None:
        self.__source = Target(json=json["source"])
        self.__sink = Target(json=json["sink"])
        self.__waypoints = [(wp[0], wp[1]) for wp in json["wayPoints"]]


class Connection:
    def __init__(self, json=None, device_ref=None):
        """[summary]

        Args:
            json ([type], optional): [description]. Defaults to None.
            device_ref ([type], optional): [description]. Defaults to None.

        Raises:
            Exception: [description]
        """
        self.name: Optional[str] = None
        self.ID: str = ""
        self.entity: Optional[str] = None
        self.params: Params = Params()
        self.source: Optional[Target] = None
        self.sinks: List[Target] = []
        self.layer: Optional[Layer] = None
        self._paths: List[ConnectionPath] = []

        if json:
            if device_ref is None:
                raise Exception(
                    "Cannot Parse Connection from JSON with no Device Reference, check device_ref parameter in constructor "
                )

            self.parse_from_json(json, device_ref)

    def parse_from_json(self, json, device_ref=None):
        """Parses from the json dict

        Args:
            json (dict): json dict after json.loads()
        """
        if device_ref is None:
            raise Exception(
                "Cannot Parse Connection from JSON with no Device Reference, check device_ref parameter in constructor "
            )

        self.name = json["name"]
        self.ID = json["id"]
        self.layer = device_ref.get_layer(json["layer"])

        # Pull out the paths
        if "paths" in json["params"].keys():
            json_paths = json["params"]["paths"]
            for json_path in json_paths:
                self._paths.append(ConnectionPath(json_path))

        self.params = Params(json["params"])

        self.source = Target(json["source"])

        if "sinks" in json.keys():
            if json["sinks"]:
                for target in json["sinks"]:
                    self.sinks.append(Target(target))
            else:
                print("connection", self.name, "does not have any sinks")
        else:
            print("connection", self.name, "does not have any sinks")

        # TODO - Change this in the v1.2 version
        if "waypoints" in json.keys():
            waypoints_raw = json["waypoints"]
            waypoints = [(wp[0], wp[1]) for wp in waypoints_raw]
            self.add_waypoints_path(None, None, waypoints)

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self.__dict__)

    @property
    def paths(self) -> List[ConnectionPath]:
        return self._paths

    def add_waypoints_path(
        self, source: Target, sink: Target, waypoints: List[Tuple[int, int]]
    ) -> None:
        """Adds a waypoints path to the connection

        Args:
            source (Target): source target of the connection corresponding to the path
            sink (Target): sink target of the connection corresponding to the path
            waypoints (List[Tuple[int, int]]): array of coordinates (as tuples)
        """
        path = ConnectionPath(source, sink, waypoints)
        self._paths.append(path)

    def to_parchmint_v1(self):
        """Returns the json dict

        Returns:
            dict: dictionary that can be used in json.dumps()
        """
        ret = {
            "sinks": [s.to_parchmint_v1() for s in self.sinks],
            "name": self.name,
            "id": self.ID,
            "source": self.source.to_parchmint_v1(),
            "params": self.params.to_parchmint_v1(),
            "layer": self.layer.ID,
        }

        ret["paths"] = [path.to_parchmint_v1() for path in self._paths]

        return ret



from typing import Optional

from parchmint.params import Params


class Layer:
    def __init__(self, json=None) -> None:
        """Creates a new instance Layer

        Args:
            json (dict, optional): json dict after json.loads(). Defaults to None.
        """
        self.ID: Optional[str] = None
        self.name: Optional[str] = None
        self.type: Optional[str] = None
        self.group: str = ""
        self.params: Params = Params()

        if json:
            self.parse_from_json(json)

    def parse_from_json(self, json):
        """Loads instance data json dict from json.loads()

        Args:
            json ([type]): [description]
        """
        self.name = json["name"]
        self.ID = json["id"]
        self.type = json["type"]
        self.group = json["group"]
        self.params = Params(json["params"])

    def to_parchmint_v1(self):
        """Returns the json dict

        Returns:
            dict: dictionary that can be used in json.dumps()
        """
        return {
            "name": self.name,
            "id": self.ID,
            "type": self.type,
            "params": self.params.to_parchmint_v1(),
            "group": self.group,
        }

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self.__dict__)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, o: object) -> bool:
        if isinstance(o, Layer):
            return o.ID == self.ID
        else:
            return False



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from parchmint.device import Device

from typing import List, Tuple

from parchmint.layer import Layer
from parchmint.params import Params
from parchmint.port import Port


class Component:
    def __init__(self, json=None, device_ref: Device = None):
        """Creates a new Component object

        Args:
            json (dict, optional): json dict after json.loads(). Defaults to None.
            device_ref (Device, optional): pointer for the Device object. Defaults to None.

        Raises:
            Exception: [description]
        """
        self.name: str = ""
        self.ID: str = ""
        self.params = Params()
        self.entity: str = ""
        self.xspan: int = -1
        self.yspan: int = -1
        self.ports: List[Port] = []
        self.layers: List[Layer] = []

        if json is not None:
            if device_ref is None:
                raise Exception(
                    "Cannot Parse Component from JSON with no Device Reference, check device_ref parameter in constructor "
                )
            self.parse_from_json(json, device_ref)

    @property
    def xpos(self) -> int:
        """returns the x coordinate of the component

        Raises:
            KeyError: when no position parameter object is found for the parchmint object

        Returns:
            int: x-coordinate
        """
        try:
            return self.params.get_param("position")[0]
        except Exception:
            print("Could not find xpos for component")
            raise KeyError

    @xpos.setter
    def xpos(self, value) -> None:
        """Sets the x-coordinate for the component

        Args:
            value (int): x coordianate of the object
        """
        if self.params.exists("position"):
            pos = self.params.get_param("position")
            pos[0] = value
            self.params.set_param("position", pos)
        else:
            self.params.set_param("position", [value, -1])

    @property
    def ypos(self) -> int:
        """Returns the y-coordinate in the parchmint object

        Raises:
            KeyError: When no position parameter is found in the parchmint object

        Returns:
            int: y coordinate of the component
        """
        try:
            return self.params.get_param("position")[1]
        except Exception:
            print("Could not find xpos for component")
            raise KeyError

    @ypos.setter
    def ypos(self, value) -> None:
        """Sets the y-coordinate of the component

        Args:
            value (int): y coordinate
        """
        if self.params.exists("position"):
            pos = self.params.get_param("position")
            pos[1] = value
            self.params.set_param("position", pos)
        else:
            self.params.set_param("position", [-1, value])

    def add_component_ports(self, ports: List[Port]) -> None:
        """Adds component ports to the component

        Args:
            ports (List[Port]): list of port objects
        """
        for port in ports:
            self.ports.append(port)

    def parse_from_json(self, json, device_ref=None):
        """Parses from the json dict

        Args:
            json (dict): json dict after json.loads()
        """
        if device_ref is None:
            raise Exception(
                "Cannot Parse Component from JSON with no Device Reference, check device_ref parameter in constructor "
            )
        self.name = json["name"]
        self.ID = json["id"]
        self.entity = json["entity"]
        self.xspan = json["x-span"]
        self.yspan = json["y-span"]
        self.params = Params(json["params"])
        self.layers = [device_ref.get_layer(layer_id) for layer_id in json["layers"]]

        for port in json["ports"]:
            self.ports.append(Port(port))

        if self.params:
            if self.params.exists("position"):
                self.xpos = self.params.get_param("position")[0]
                self.ypos = self.params.get_param("position")[1]

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self.__dict__)

    def to_parchmint_v1(self):
        """Returns the json dict

        Returns:
            dict: dictionary that can be used in json.dumps()
        """
        ret = {
            "name": self.name,
            "id": self.ID,
            "layers": [layer.ID for layer in self.layers],
            "params": self.params.to_parchmint_v1(),
            "ports": [p.to_parchmint_v1() for p in self.ports],
            "entity": self.entity,
            "x-span": int(self.xspan),
            "y-span": int(self.yspan),
        }

        return ret

    def __eq__(self, obj):
        if isinstance(obj, Component):
            return obj.ID == self.ID
        else:
            return False

    def get_port(self, label: str) -> Port:
        """Returns a port in the component identified by the corresponding label

        Args:
            label (str): label of the componentport

        Raises:
            Exception: if there is no component port with the corresponding label is found

        Returns:
            Port: component port
        """
        for port in self.ports:
            if label == port.label:
                return port
        raise Exception("Could not find port with the label: {}".format(label))

    def get_absolute_port_coordinates(self, port_label: str) -> Tuple[float, float]:
        """Gets the absolute coordinates of the component port identified by the label

        Args:
            port_label (str): unique identifier for the component port

        Returns:
            Tuple[float, float]: coordinates of the component port
        """
        port = self.get_port(port_label)
        x = self.xpos + port.x
        y = self.ypos + port.y
        return (x, y)
